@@ -122,6 +122,14 @@ DevToolsAPIImpl.prototype = {
         this._dispatchOnInspectorFrontendAPI("devicesDiscoveryConfigChanged", [discoverUsbDevices, portForwardingEnabled, portForwardingConfig]);
     },
 
+    /**
+     * @param {!Adb.PortForwardingStatus} status
+     */
+    devicesPortForwardingStatusChanged: function(status)
+    {
+        this._dispatchOnInspectorFrontendAPI("devicesPortForwardingStatusChanged", [status]);
+    },
+
     /**
      * @param {!Array.<!Adb.Device>} devices
      */
@@ -152,6 +160,15 @@ DevToolsAPIImpl.prototype = {
         this._dispatchOnInspectorFrontendAPI("enterInspectElementMode", []);
     },
 
+    /**
+     * @param {number} callId
+     * @param {string} script
+     */
+    evaluateForTestInFrontend: function(callId, script)
+    {
+        this._dispatchOnInspectorFrontendAPI("evaluateForTestInFrontend", [callId, script]);
+    },
+
     /**
      * @param {!Array.<!{fileSystemName: string, rootURL: string, fileSystemPath: string}>} fileSystems
      */
@@ -289,24 +306,6 @@ DevToolsAPIImpl.prototype = {
     streamWrite: function(id, chunk)
     {
         this._dispatchOnInspectorFrontendAPI("streamWrite", [id, chunk]);
-    },
-
-    frontendAPIAttached: function()
-    {
-        this._dispatchOnInspectorFrontendAPI("frontendAPIAttached", []);
-    },
-
-    frontendAPIDetached: function()
-    {
-        this._dispatchOnInspectorFrontendAPI("frontendAPIDetached", []);
-    },
-
-    /**
-     * @param {string} command
-     */
-    dispatchFrontendAPIMessage: function(command)
-    {
-        this._dispatchOnInspectorFrontendAPI("dispatchFrontendAPIMessage", [command]);
     }
 }
 
@@ -526,18 +525,12 @@ InspectorFrontendHostImpl.prototype = {
      */
     recordEnumeratedHistogram: function(actionName, actionCode, bucketSize)
     {
+        // Support for M49 frontend.
+        if (actionName === "DevTools.DrawerShown")
+            return;
         DevToolsAPI.sendMessageToEmbedder("recordEnumeratedHistogram", [actionName, actionCode, bucketSize], null);
     },
 
-    /**
-     * @override
-     * @param {string} message
-     */
-    sendFrontendAPINotification: function(message)
-    {
-        DevToolsAPI.sendMessageToEmbedder("sendFrontendAPINotification", [message], null);
-    },
-
     /**
      * @override
      */
@@ -665,6 +658,14 @@ InspectorFrontendHostImpl.prototype = {
         return DevToolsHost.isUnderTest();
     },
 
+    /**
+     * @override
+     */
+    readyForTest: function()
+    {
+        DevToolsAPI.sendMessageToEmbedder("readyForTest", [], null);
+    },
+
     /**
      * @override
      * @param {boolean} discoverUsbDevices
@@ -728,6 +729,14 @@ InspectorFrontendHostImpl.prototype = {
 
     // Backward-compatible methods below this line --------------------------------------------
 
+    /**
+     * Support for legacy front-ends (<M50).
+     * @param {string} message
+     */
+    sendFrontendAPINotification: function(message)
+    {
+    },
+
     /**
      * Support for legacy front-ends (<M41).
      * @return {string}
@@ -853,6 +862,105 @@ window.InspectorFrontendHost = new InspectorFrontendHostImpl();
 
 // DevToolsApp ---------------------------------------------------------------
 
+function installObjectObserve()
+{
+    var properties = [
+        "advancedSearchConfig", "auditsPanelSplitViewState", "auditsSidebarWidth", "blockedURLs", "breakpoints", "cacheDisabled", "colorFormat", "consoleHistory",
+        "consoleTimestampsEnabled", "cpuProfilerView", "cssSourceMapsEnabled", "currentDockState", "customColorPalette", "customDevicePresets", "customEmulatedDeviceList",
+        "customFormatters", "customUserAgent", "databaseTableViewVisibleColumns", "dataGrid-cookiesTable", "dataGrid-DOMStorageItemsView", "debuggerSidebarHidden", "disableDataSaverInfobar",
+        "disablePausedStateOverlay", "domBreakpoints", "domWordWrap", "elementsPanelSplitViewState", "elementsSidebarWidth", "emulation.deviceHeight", "emulation.deviceModeValue",
+        "emulation.deviceOrientationOverride", "emulation.deviceScale", "emulation.deviceScaleFactor", "emulation.deviceUA", "emulation.deviceWidth", "emulation.geolocationOverride",
+        "emulation.showDeviceMode", "emulation.showRulers", "enableAsyncStackTraces", "eventListenerBreakpoints", "fileMappingEntries", "fileSystemMapping", "FileSystemViewSidebarWidth",
+        "fileSystemViewSplitViewState", "filterBar-consoleView", "filterBar-networkPanel", "filterBar-promisePane", "filterBar-timelinePanel", "frameViewerHideChromeWindow",
+        "heapSnapshotRetainersViewSize", "heapSnapshotSplitViewState", "hideCollectedPromises", "hideNetworkMessages", "highlightNodeOnHoverInOverlay", "highResolutionCpuProfiling",
+        "inlineVariableValues", "Inspector.drawerSplitView", "Inspector.drawerSplitViewState", "InspectorView.panelOrder", "InspectorView.screencastSplitView",
+        "InspectorView.screencastSplitViewState", "InspectorView.splitView", "InspectorView.splitViewState", "javaScriptDisabled", "jsSourceMapsEnabled", "lastActivePanel", "lastDockState",
+        "lastSelectedSourcesSidebarPaneTab", "lastSnippetEvaluationIndex", "layerDetailsSplitView", "layerDetailsSplitViewState", "layersPanelSplitViewState", "layersShowInternalLayers",
+        "layersSidebarWidth", "messageLevelFilters", "messageURLFilters", "monitoringXHREnabled", "navigatorGroupByFolder", "navigatorHidden", "networkColorCodeResourceTypes",
+        "networkConditions", "networkConditionsCustomProfiles", "networkHideDataURL", "networkLogColumnsVisibility", "networkLogLargeRows", "networkLogShowOverview",
+        "networkPanelSplitViewState", "networkRecordFilmStripSetting", "networkResourceTypeFilters", "networkShowPrimaryLoadWaterfall", "networkSidebarWidth", "openLinkHandler",
+        "pauseOnCaughtException", "pauseOnExceptionEnabled", "preserveConsoleLog", "prettyPrintInfobarDisabled", "previouslyViewedFiles", "profilesPanelSplitViewState",
+        "profilesSidebarWidth", "promiseStatusFilters", "recordAllocationStacks", "requestHeaderFilterSetting", "request-info-formData-category-expanded",
+        "request-info-general-category-expanded", "request-info-queryString-category-expanded", "request-info-requestHeaders-category-expanded",
+        "request-info-requestPayload-category-expanded", "request-info-responseHeaders-category-expanded", "resources", "resourcesLastSelectedItem", "resourcesPanelSplitViewState",
+        "resourcesSidebarWidth", "resourceViewTab", "savedURLs", "screencastEnabled", "scriptsPanelNavigatorSidebarWidth", "searchInContentScripts", "selectedAuditCategories",
+        "selectedColorPalette", "selectedProfileType", "shortcutPanelSwitch", "showAdvancedHeapSnapshotProperties", "showEventListenersForAncestors", "showFrameowkrListeners",
+        "showHeaSnapshotObjectsHiddenProperties", "showInheritedComputedStyleProperties", "showMediaQueryInspector", "showNativeFunctionsInJSProfile", "showUAShadowDOM",
+        "showWhitespacesInEditor", "sidebarPosition", "skipContentScripts", "skipStackFramesPattern", "sourceMapInfobarDisabled", "sourcesPanelDebuggerSidebarSplitViewState",
+        "sourcesPanelNavigatorSplitViewState", "sourcesPanelSplitSidebarRatio", "sourcesPanelSplitViewState", "sourcesSidebarWidth", "standardEmulatedDeviceList",
+        "StylesPaneSplitRatio", "stylesPaneSplitViewState", "textEditorAutocompletion", "textEditorAutoDetectIndent", "textEditorBracketMatching", "textEditorIndent",
+        "timelineCaptureFilmStrip", "timelineCaptureLayersAndPictures", "timelineCaptureMemory", "timelineCaptureNetwork", "timeline-details", "timelineEnableJSSampling",
+        "timelineOverviewMode", "timelinePanelDetailsSplitViewState", "timelinePanelRecorsSplitViewState", "timelinePanelTimelineStackSplitViewState", "timelinePerspective",
+        "timeline-split", "timelineTreeGroupBy", "timeline-view", "timelineViewMode", "uiTheme", "watchExpressions", "WebInspector.Drawer.lastSelectedView", "WebInspector.Drawer.showOnLoad",
+        "workspaceExcludedFolders", "workspaceFolderExcludePattern", "workspaceInfobarDisabled", "workspaceMappingInfobarDisabled", "xhrBreakpoints"];
+
+    /**
+     * @this {!{_storage: Object, _name: string}}
+     */
+    function settingRemove()
+    {
+        this._storage[this._name] = undefined;
+    }
+
+    function objectObserve(object, observer)
+    {
+        if (window["WebInspector"]) {
+            var settingPrototype = window["WebInspector"]["Setting"]["prototype"];
+            if (typeof settingPrototype["remove"] === "function")
+                settingPrototype["remove"] = settingRemove;
+        }
+
+        var changedProperties = new Set();
+        var scheduled = false;
+
+        function scheduleObserver()
+        {
+            if (!scheduled) {
+                scheduled = true;
+                setImmediate(callObserver);
+            }
+        }
+
+        function callObserver()
+        {
+            scheduled = false;
+            var changes = [];
+            changedProperties.forEach(function(name) { changes.push({name: name}); });
+            changedProperties.clear();
+            observer.call(null, changes);
+        }
+
+        var storage = new Map();
+
+        function defineProperty(property)
+        {
+            if (property in object) {
+                storage.set(property, object[property]);
+                delete object[property];
+            }
+
+            Object.defineProperty(object, property, {
+                get: function()
+                {
+                    return storage.get(property);
+                },
+
+                set: function(value)
+                {
+                    storage.set(property, value);
+                    changedProperties.add(property);
+                    scheduleObserver();
+                }
+            });
+        }
+
+        for (var i = 0; i < properties.length; ++i)
+            defineProperty(properties[i]);
+    }
+
+    window.Object.observe = objectObserve;
+}
+
 /**
  * @suppressGlobalPropertiesCheck
  */
@@ -861,6 +969,9 @@ function installBackwardsCompatibility()
     if (window.location.search.indexOf("remoteFrontend") === -1)
         return;
 
+    // Support for legacy (<M50) frontends.
+    installObjectObserve();
+
     /**
      * @this {CSSStyleDeclaration}
      */
@@ -893,7 +1004,15 @@ function installBackwardsCompatibility()
     var styleElement = window.document.createElement("style");
     styleElement.type = "text/css";
     styleElement.textContent = "html /deep/ * { min-width: 0; min-height: 0; }";
+
+    // Support for quirky border-image behavior (<M51), see:
+    // https://bugs.chromium.org/p/chromium/issues/detail?id=559258
+    styleElement.textContent += "\nhtml /deep/ .cm-breakpoint .CodeMirror-linenumber { border-style: solid !important; }";
+    styleElement.textContent += "\nhtml /deep/ .cm-breakpoint.cm-breakpoint-conditional .CodeMirror-linenumber { border-style: solid !important; }";
     window.document.head.appendChild(styleElement);
+
+    // Support for legacy (<M49) frontends. Remove in M52.
+    Event.prototype.deepPath = undefined;
 }
 
 function windowLoaded()
@@ -907,33 +1026,14 @@ if (window.document.head && (window.document.readyState === "complete" || window
 else
     window.addEventListener("DOMContentLoaded", windowLoaded, false);
 
-// UITests ------------------------------------------------------------------
-
-if (window.domAutomationController) {
-    var uiTests = {};
-
-    uiTests._dispatchIfReady = function()
-    {
-        if (uiTests._testSuite && uiTests._pendingDispatchArgs) {
-            var args = uiTests._pendingDispatchArgs;
-            delete uiTests._pendingDispatchArgs;
-            uiTests._testSuite.dispatch(args);
-        }
-    }
-
-    uiTests.dispatchOnTestSuite = function(args)
-    {
-        uiTests._pendingDispatchArgs = args;
-        uiTests._dispatchIfReady();
-    };
-
-    uiTests.testSuiteReady = function(testSuiteConstructor)
-    {
-        uiTests._testSuite = testSuiteConstructor(window.domAutomationController);
-        uiTests._dispatchIfReady();
-    };
-
-    window.uiTests = uiTests;
-}
-
 })(window);
+
+if (!DOMTokenList.prototype.__originalDOMTokenListToggle) {
+    DOMTokenList.prototype.__originalDOMTokenListToggle = DOMTokenList.prototype.toggle;
+    DOMTokenList.prototype.toggle = function(token, force)
+    {
+        if (arguments.length === 1)
+            force = !this.contains(token);
+        return this.__originalDOMTokenListToggle(token, !!force);
+    }
+}
